﻿using System;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using RootPro.RootProCAD;
using RootPro.RootProCAD.UI;
using RootPro.RootProCAD.Geometry;
using System.Drawing;

///////////////////////////////////////////////////////////////////////////////////////////////////
// TabViewのWindowsイベントをフックするNativeWindowクラス
// Copyright @ 2019 RootPro Co.,Ltd. 
///////////////////////////////////////////////////////////////////////////////////////////////////
namespace RCAddInAdvancedZoomCS
{

    class RCTabViewNativeWindow : NativeWindow
    {
        [DllImport("user32.dll", EntryPoint = "SendMessageA")]
        extern static int SendMessage(IntPtr hWnd, Int32 Msg, IntPtr wParam, IntPtr lParam);

        [DllImport("user32.dll")]
        extern static bool ClientToScreen(IntPtr hWnd, ref Point lpPoint);

        private const int WM_MOUSEMOVE = 0x200;
        private const int WM_LBUTTONDOWN = 0x201;
        private const int WM_LBUTTONUP = 0x202;
        private const int WM_RBUTTONDOWN = 0x204;
        private const int WM_RBUTTONUP = 0x205;
        private const int WM_MBUTTONDOWN = 0x207;
        private const int WM_MBUTTONUP = 0x208;

        /// <summary>
        /// メッセージ監視
        /// </summary>
        private RCMouseEventMonitor mouseEventMonitor;

        /// <summary>
        /// AppAddIn
        /// </summary>
        private AppAddIn app;


        /// <summary>
        /// 前回表示の左下座標
        /// </summary>
        private Point2d viewBottomLeftPoint;

        /// <summary>
        /// 前回表示の右上座標
        /// </summary>
        private Point2d viewTopRightPoint;

        private enum FunctionType
        {
            Nothing = 0,                // なし
            ScrollTo = 1,               // 中心に表示
            ZoomAll = 2,                // 全図形表示
            ZoomRange = 3,              // 範囲拡大
            LastView = 4,               // 前倍率
            ZoomOut = 5,                // 縮小表示
        }

        private FunctionType functionType;

        /// <summary>
        /// コンストラクタ
        /// </summary>
        public RCTabViewNativeWindow(AppAddIn app)
        {
            this.app = app;
            mouseEventMonitor = new RCMouseEventMonitor();
            functionType = FunctionType.Nothing;

            viewBottomLeftPoint = app.Geometry.CreatePoint(0.0, 0.0);
            viewTopRightPoint = app.Geometry.CreatePoint(0.0, 0.0);
        }

        protected override void WndProc(ref System.Windows.Forms.Message m)
        {
            // メッセージを処理するかチェック
            if (mouseEventMonitor.TestMessage(m))
            {
                // メッセージの変換
                RCMouseEventMonitor.FilterResult result = mouseEventMonitor.FilterMessage(ref m);

                // 状態なし
                if (mouseEventMonitor.mouseState == MouseState.Nothing)
                {
                    if (this.functionType != FunctionType.Nothing)
                    {
                        // 状態が変わった場合はステータスバークリア
                        this.functionType = FunctionType.Nothing;
                        app.StatusBar.SetInformation("", 0);
                    }
                }
                else
                {
                    // LR クリック後の場合
                    if ((mouseEventMonitor.mouseState & MouseState.LRButtonDown) == MouseState.LRButtonDown)
                    {
                        // 対応する機能タイプを取得する
                        FunctionType currentFunctionType = GetCurrentFunctionType(m);
                        if(this.functionType != currentFunctionType)
                        {
                            String str = "";
                            switch(currentFunctionType)
                            {
                                // なし
                                case FunctionType.Nothing:
                                    str = "";
                                    break;
                                // ビューの中心に表示
                                case FunctionType.ScrollTo:
                                    str = Properties.Resources.FunctionScrollTo;
                                    break;
                                // 全図形
                                case FunctionType.ZoomAll:
                                    str = Properties.Resources.FunctionZoomAll;
                                    break;
                                // 二点拡大
                                case FunctionType.ZoomRange:
                                    str = Properties.Resources.FunctionZoomRange;
                                    break;
                                // 前回のビュー表示
                                case FunctionType.LastView:
                                    str = Properties.Resources.FunctionLastView;
                                    break;
                                // 縮小
                                case FunctionType.ZoomOut:
                                    str = Properties.Resources.FunctionZoomOut;
                                    break;

                            }
                            // ステータスバーを更新
                            app.StatusBar.SetInformation(str, 0);
                            this.functionType = currentFunctionType;
                        }
                    }
                }

                // メッセージ無変換、または変更のみの場合
                if (result == RCMouseEventMonitor.FilterResult.NoChange ||
                    result == RCMouseEventMonitor.FilterResult.MessageChanged)
                {
                    base.WndProc(ref m);
                    return;
                }

                // メッセージ蓄積、またはストップ
                else if (result == RCMouseEventMonitor.FilterResult.MessageStored ||
                    result == RCMouseEventMonitor.FilterResult.MessageStop)
                {
                    return;
                }
                // 蓄積メッセージあり
                else if (result == RCMouseEventMonitor.FilterResult.ExistStoredMessages)
                {
                    // 蓄積されたメッセージを先に送信する
                    for (int i = 0; i < mouseEventMonitor.GetStoredMsgCount(); i++)
                    {
                        Message storedMessage = mouseEventMonitor.GetStoredMsg(i);
                        SendMessage(storedMessage.HWnd, storedMessage.Msg, storedMessage.WParam, storedMessage.LParam);
                    }
                    base.WndProc(ref m);

                    mouseEventMonitor.ClearState();
                    return;
                }
                // 機能の実行
                else if (result == RCMouseEventMonitor.FilterResult.ExecuteFunction)
                {
                    ExecuteFunction();
                    return;
                }
            }
            base.WndProc(ref m);
        }

        /// <summary>
        /// 前回表示の情報をクリア
        /// </summary>
        public void ResetLastViewPoint()
        {
            viewBottomLeftPoint.X = 0.0;
            viewBottomLeftPoint.Y = 0.0;
            viewTopRightPoint.X = 0.0;
            viewTopRightPoint.Y = 0.0;
        }

        /// <summary>
        /// 機能の実行
        /// </summary>
        private bool ExecuteFunction()
        {
            if (mouseEventMonitor.GetStoredMsgCount() != 2)
                return false;

            // 部分図ビューではない場合はリターン
            if (!(app.ActiveTabView is DrawingView))
                return false;

            Message msg = mouseEventMonitor.GetStoredMsg(0);
            // LParam の上位2バイトを取得
            uint lp = unchecked(IntPtr.Size == 8 ? (uint)msg.LParam.ToInt64() : (uint)msg.LParam.ToInt32());
            int upperLParam1 = (short)(lp >> 16);
            int lowLParam1 = (int)(lp & 0x0000ffff);

            msg = mouseEventMonitor.GetStoredMsg(1);
            // LParam の上位2バイトを取得
            lp = unchecked(IntPtr.Size == 8 ? (uint)msg.LParam.ToInt64() : (uint)msg.LParam.ToInt32());
            int upperLParam2 = (short)(lp >> 16);
            int lowLParam2 = (int)(lp & 0x0000ffff);

            // 部分図ビューを取得
            DrawingView drawingView = (DrawingView)app.ActiveTabView;

            // 1点目の座標をクライアント座標からスクリーン座標に変換
            Point point = new Point(lowLParam1, upperLParam1);
            ClientToScreen(msg.HWnd, ref point);

            // スクリーン座標から部分図ビューのワールド座標に変換
            Point2d screenPoint = app.Geometry.CreatePoint(point.X, point.Y);
            Point2d worldPoint = drawingView.TransformPointScreenToWorld(screenPoint);

            // 2点目の座標をクライアント座標からスクリーン座標に変換
            Point point2 = new Point(lowLParam2, upperLParam2);
            ClientToScreen(msg.HWnd, ref point2);

            // スクリーン座標から部分図ビューのワールド座標に変換
            Point2d screenPoint2 = app.Geometry.CreatePoint(point2.X, point2.Y);
            Point2d worldPoint2 = drawingView.TransformPointScreenToWorld(screenPoint2);

            // 前回表示の範囲をとっておく
            Point2d lastBottomLeft = this.viewBottomLeftPoint;
            Point2d lastTopRight = this.viewTopRightPoint;

            // 現在の表示に置き換える
            this.viewBottomLeftPoint = drawingView.ViewBottomLeftPoint;
            this.viewTopRightPoint = drawingView.ViewTopRightPoint;

            switch (this.functionType)
            {
                // ビューの中心に表示
                case FunctionType.ScrollTo:
                    drawingView.ScrollTo(worldPoint);
                    break;
                // 全図形
                case FunctionType.ZoomAll:
                    drawingView.ZoomAll();
                    break;
                // 二点拡大
                case FunctionType.ZoomRange:
                    drawingView.ZoomWindow(worldPoint, worldPoint2);
                    break;
                // 前回のビュー表示
                case FunctionType.LastView:
                    if (!lastBottomLeft.Equals(lastTopRight))
                        drawingView.ZoomWindow(lastBottomLeft, lastTopRight);
                    break;
                // 縮小
                case FunctionType.ZoomOut:
                    {
                        // 左上移動は縮小
                        // マウスの一点目を中心にして、範囲を倍にする
                        double cx = (drawingView.ViewTopRightPoint.X - drawingView.ViewBottomLeftPoint.X);
                        double cy = (drawingView.ViewTopRightPoint.Y - drawingView.ViewBottomLeftPoint.Y);

                        Point2d pBottomLeft = app.Geometry.CreatePoint(
                            worldPoint.X - cx, worldPoint.Y - cy);
                        Point2d pTopRight = app.Geometry.CreatePoint(
                            worldPoint.X + cx, worldPoint.Y + cy);

                        drawingView.ZoomWindow(pBottomLeft, pTopRight);
                    }
                    break;
                case FunctionType.Nothing:
                    break;
                default:
                    break;

            }

            // ステータスバーの表示をクリア
            app.StatusBar.SetInformation("", 0);

            return true;
        }

        /// <summary>
        /// 現在のマウス位置に対応する機能タイプを取得
        /// </summary>
        private FunctionType GetCurrentFunctionType(Message m)
        {

            if (mouseEventMonitor.GetStoredMsgCount() == 0)
                return FunctionType.Nothing;

            if (!(app.ActiveTabView is DrawingView))
                return FunctionType.Nothing;

            int upperLParam1 = 0, lowLParam1 = 0, upperLParam2 = 0, lowLParam2 = 0;

            Message msg = mouseEventMonitor.GetStoredMsg(0);
            // LParam の上位2バイト、下位2バイトを取得
            uint lp = unchecked(IntPtr.Size == 8 ? (uint)msg.LParam.ToInt64() : (uint)msg.LParam.ToInt32());
            upperLParam1 = (short)(lp >> 16);
            lowLParam1 = (int)(lp & 0x0000ffff);

            if (mouseEventMonitor.GetStoredMsgCount() > 1)
                msg = mouseEventMonitor.GetStoredMsg(1);
            else
                msg = m;

            // LParam の上位2バイト、下位2バイトを取得
            lp = unchecked(IntPtr.Size == 8 ? (uint)msg.LParam.ToInt64() : (uint)msg.LParam.ToInt32());
            upperLParam2 = (short)(lp >> 16);
            lowLParam2 = (int)(lp & 0x0000ffff);

            Point point = new Point(lowLParam1, upperLParam1);
            ClientToScreen(msg.HWnd, ref point);

            Point2d screenPoint = app.Geometry.CreatePoint(point.X, point.Y);

            Point point2 = new Point(lowLParam2, upperLParam2);
            ClientToScreen(msg.HWnd, ref point2);
            Point2d screenPoint2 = app.Geometry.CreatePoint(point2.X, point2.Y);

            // マウスが移動しなかった
            if (Math.Abs(lowLParam1 - lowLParam2) < 10 && Math.Abs(upperLParam1 - upperLParam2) < 10)
            {
                return FunctionType.ScrollTo;
            }
            else
            {
                // 右方向に移動
                if (lowLParam1 < lowLParam2)
                {
                    // 下方向に移動
                    if (upperLParam1 < upperLParam2)
                    {
                        // 右下移動は二点拡大
                        return FunctionType.ZoomRange;
                    }
                    // 上方向に移動
                    else
                    {
                        // 右上移動は全図形表示
                        return FunctionType.ZoomAll;
                    }
                }
                // 左方向に移動
                else
                {
                    // 下方向に移動
                    if (upperLParam1 < upperLParam2)
                    {
                        // 右上移動は全図形表示
                        return FunctionType.LastView;
                    }
                    // 上方向に移動
                    else
                    {
                        // 左上移動は縮小
                        return FunctionType.ZoomOut;
                    }
                }
            }
        }
    }
}