﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows.Forms;

namespace RCAddInAdvancedZoomCS
{
    /// <summary>
    /// 状態フラグ
    /// </summary>
    [Flags]
    public enum MouseState
    {
        Nothing = 0x000,        // マウス状態なし
        LButtonDown = 0x001,    // 左クリックDOWN
        RButtonDown = 0x002,    // 右クリックDOWN
        Moved = 0x004,          // マウス移動
        LButtonUp = 0x008,      // 左クリックUP
        RButtonUp = 0x010,      // 右クリックUP

        LRButtonDown = LButtonDown | RButtonDown,
        LRAndLUp = LButtonDown | RButtonDown | LButtonUp,
        LRAndRUp = LButtonDown | RButtonDown | RButtonUp,
        LRMoving = LButtonDown | RButtonDown | Moved,
        LRMovingAndLUp = LButtonDown | RButtonDown | Moved | LButtonUp,
        LRMovingAndRUp = LButtonDown | RButtonDown | Moved | RButtonUp,
    }

    class RCMouseEventMonitor
    {

        private const int WM_MOUSEMOVE = 0x200;
        private const int WM_LBUTTONDOWN = 0x201;
        private const int WM_LBUTTONUP = 0x202;
        private const int WM_RBUTTONDOWN = 0x204;
        private const int WM_RBUTTONUP = 0x205;
        private const int WM_MBUTTONDOWN = 0x207;
        private const int WM_MBUTTONUP = 0x208;
        private const int WM_MOUSELEAVE = 0x2A3;
        private const int MK_CONTROL = 0x008;

        /// <summary>
        /// 蓄積されたメッセージのリスト
        /// </summary>
        private List<Message> storedMsgList;

        /// <summary>
        /// 状態
        /// </summary>
        public MouseState mouseState;

        /// <summary>
        /// フィルター結果
        /// </summary>
        public enum FilterResult
        {
            NoChange = 0,               // メッセージ変更なし
            MessageChanged = 1,         // メッセージが変更された
            MessageStored = 2,          // メッセージ蓄積
            MessageStop = 3,            // メッセージ中止
            ExistStoredMessages = 4,    // 蓄積メッセージあり
            ExecuteFunction = 5,        // 機能の実行
        }

        /// <summary>
        /// Ctrl キーが押されているか
        /// </summary>
        private bool ctrlState;

        /// <summary>
        /// コンストラクタ
        /// </summary>
        public RCMouseEventMonitor()
        {
            storedMsgList = new List<Message>();

            mouseState = MouseState.Nothing;
            ctrlState = false;
        }

        /// <summary>
        /// メッセージを処理するかチェックする
        /// </summary>
        public bool TestMessage(Message m)
        {
            switch(m.Msg)
            {
                case WM_LBUTTONDOWN:
                case WM_RBUTTONDOWN:
                case WM_LBUTTONUP:
                case WM_RBUTTONUP:
                case WM_MOUSEMOVE:
                case WM_MOUSELEAVE:
                    return true;
                default:
                    break;
            }

            return false;
        }

        /// <summary>
        /// メッセージを変換する
        /// </summary>
        public FilterResult FilterMessage(ref Message msg)
        {
            FilterResult filterResult = FilterResult.NoChange;
            switch (msg.Msg)
            {
                // 左クリックダウン
                case WM_LBUTTONDOWN:
                    if (mouseState == MouseState.Nothing)
                    {
                        ctrlState = GetCtrlKeyState(msg);

                        filterResult = FilterResult.MessageStored;
                        mouseState = MouseState.LButtonDown;

                        storedMsgList.Add(msg);
                    }
                    else if (mouseState == MouseState.RButtonDown)
                    {
                        if (ctrlState == GetCtrlKeyState(msg))
                        {
                            // Ctrl キーを押している場合はパン
                            if (ctrlState == true)
                            {
                                storedMsgList.Clear();

                                msg.Msg = WM_MBUTTONDOWN;
                                filterResult = FilterResult.MessageChanged;
                                mouseState |= MouseState.LButtonDown;
                            }
                            else
                            {
                                filterResult = FilterResult.MessageStop;
                                mouseState |= MouseState.LButtonDown;
                            }
                        }
                        else
                        {
                            filterResult = FilterResult.NoChange;
                            ClearState();
                        }
                    }
                    else
                    {
                        filterResult = FilterResult.NoChange;
                        ClearState();
                    }
                    break;
                // 右クリックダウン
                case WM_RBUTTONDOWN:
                    if (mouseState == MouseState.Nothing)
                    {
                        ctrlState = GetCtrlKeyState(msg);

                        filterResult = FilterResult.MessageStored;
                        mouseState = MouseState.RButtonDown;

                        storedMsgList.Add(msg);
                    }
                    else if (mouseState == MouseState.LButtonDown)
                    {
                        if (ctrlState == GetCtrlKeyState(msg))
                        {
                            // Ctrl キーを押している場合はパン
                            if (ctrlState == true)
                            {
                                storedMsgList.Clear();

                                msg.Msg = WM_MBUTTONDOWN;
                                filterResult = FilterResult.MessageChanged;
                                mouseState |= MouseState.RButtonDown;
                            }
                            else
                            {
                                filterResult = FilterResult.MessageStop;
                                mouseState |= MouseState.RButtonDown;
                            }
                        }
                        else
                        {
                            filterResult = FilterResult.NoChange;
                            ClearState();
                        }
                    }
                    else
                    {
                        filterResult = FilterResult.NoChange;
                        ClearState();
                    }
                    break;
                // 左クリックアップ
                case WM_LBUTTONUP:
                    if (mouseState == MouseState.LRMoving ||
                        mouseState == MouseState.LRButtonDown)
                    {
                        if (ctrlState == GetCtrlKeyState(msg))
                        {
                            // Ctrl キーを押している場合はパン終了
                            if (ctrlState == true)
                            {
                                msg.Msg = WM_MBUTTONUP;
                                filterResult = FilterResult.MessageChanged;
                                mouseState |= MouseState.LButtonUp;
                            }
                            else
                            {
                                storedMsgList.Add(msg);

                                filterResult = FilterResult.ExecuteFunction;
                                mouseState |= MouseState.LButtonUp;
                            }
                        }
                        else
                        {
                            filterResult = FilterResult.NoChange;
                            ClearState();
                        }
                    }
                    else if (mouseState == MouseState.LRMovingAndRUp ||
                        mouseState == MouseState.LRAndRUp)
                    {
                        filterResult = FilterResult.MessageStop;
                        ClearState();
                    }
                    else if (mouseState == MouseState.LButtonDown)
                    {
                        filterResult = FilterResult.ExistStoredMessages;
                    }
                    else
                    {
                        filterResult = FilterResult.NoChange;
                        ClearState();
                    }
                    break;
                // 右クリックアップ
                case WM_RBUTTONUP:
                    if (mouseState == MouseState.LRMoving ||
                        mouseState == MouseState.LRButtonDown)
                    {

                        if (ctrlState == GetCtrlKeyState(msg))
                        {
                            // Ctrl キーを押している場合はパン終了
                            if (ctrlState == true)
                            {
                                msg.Msg = WM_MBUTTONUP;
                                filterResult = FilterResult.MessageChanged;
                                mouseState |= MouseState.RButtonUp;
                            }
                            else
                            {
                                storedMsgList.Add(msg);

                                filterResult = FilterResult.ExecuteFunction;
                                mouseState |= MouseState.RButtonUp;
                            }
                        }
                        else
                        {
                            filterResult = FilterResult.NoChange;
                            ClearState();
                        }
                    }
                    else if (mouseState == MouseState.LRMovingAndLUp ||
                        mouseState == MouseState.LRAndLUp)
                    {
                        filterResult = FilterResult.MessageStop;
                        ClearState();
                    }
                    else
                    {
                        filterResult = FilterResult.NoChange;
                        ClearState();
                    }
                    break;
                // マウス移動
                case WM_MOUSEMOVE:
                    if (mouseState == MouseState.LRButtonDown ||
                        mouseState == MouseState.LRMoving)
                    {
                        if (ctrlState == GetCtrlKeyState(msg))
                        {
                            filterResult = FilterResult.NoChange;

                            if((mouseState & MouseState.Moved) == 0)
                                mouseState |= MouseState.Moved;
                        }
                        // パンの中止
                        else if(ctrlState == true)
                        {
                            msg.Msg = WM_MBUTTONUP;
                            filterResult = FilterResult.MessageChanged;
                            ClearState();
                        }
                        // 機能の中止
                        else
                        {
                            filterResult = FilterResult.MessageStop;
                            ClearState();
                        }
                    }
                    else if(mouseState == MouseState.LButtonDown ||
                        mouseState == MouseState.RButtonDown)
                    {
                        filterResult = FilterResult.ExistStoredMessages;
                    }
                    break;
                // マウス移動マウスが図面から離れた
                case WM_MOUSELEAVE:
                    filterResult = FilterResult.NoChange;
                    ClearState();
                    break;
            default:
                break;
            }

            return filterResult;
        }

        /// <summary>
        /// 蓄積メッセージの個数を取得する
        /// </summary>
        public int GetStoredMsgCount()
        {
            return storedMsgList.Count();
        }

        /// <summary>
        /// 指定したインデックスの事前に送るメッセージを取得する
        /// </summary>
        public Message GetStoredMsg(int index)
        {
            return storedMsgList[index];
        }
 
        /// <summary>
        /// 状態をクリアする
        /// </summary>
        public void ClearState()
        {
            mouseState = MouseState.Nothing;
            ctrlState = false;

            storedMsgList.Clear();
        }

        /// <summary>
        /// Ctrl キーが押されているかチェック
        /// </summary>
        private bool GetCtrlKeyState(Message m)
        {
            switch (m.Msg)
            {
                case WM_LBUTTONDOWN:
                case WM_RBUTTONDOWN:
                case WM_LBUTTONUP:
                case WM_RBUTTONUP:
                case WM_MOUSEMOVE:
                    return (((int)m.WParam & MK_CONTROL) != 0) ? true : false;
                default:
                    break;
            }
            return false;
        }
    }
}
